//
//  ContentView.swift
//  USB Meter
//
//  Created by Bogdan Timofte on 01/03/2020.
//  Copyright © 2020 Bogdan Timofte. All rights reserved.
//

//MARK: Bluetooth Icon: https://upload.wikimedia.org/wikipedia/commons/d/da/Bluetooth.svg

import SwiftUI

struct ContentView: View {
    var body: some View {
        NavigationView {
            SidebarView()
                .navigationBarTitle(Text("USB Meters"), displayMode: .inline)
        }
    }
}
