//
//  MeterCurrentScreenSummaryView.swift
//  USB Meter
//
//  Created by Bogdan Timofte on 29/03/2026.
//  Co-authored-by: GPT-5.3-Codex.
//  Copyright © 2026 Bogdan Timofte. All rights reserved.
//

import SwiftUI

struct MeterCurrentScreenSummaryView: View {
    let reportsCurrentScreenIndex: Bool
    let currentScreenDescription: String
    let isExpandedCompactLayout: Bool

    var body: some View {
        if reportsCurrentScreenIndex {
            Text(currentScreenDescription)
                .font((isExpandedCompactLayout ? Font.title3 : .subheadline).weight(.semibold))
                .multilineTextAlignment(.center)
        } else {
            VStack {
                Image(systemName: "questionmark.square.dashed")
                    .font(.system(size: isExpandedCompactLayout ? 30 : 24, weight: .semibold))
                    .foregroundColor(.secondary)
            }
        }
    }
}